/* 
   SquidTL - Time Limit for Squid proxy
   Copyright (C) 2009-2010, Michele Pinassi <o-zone@zerozone.it>.
   
   http://www.zerozone.it/Software/Linux/SquidTL
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "HTEscape.h"
#include <getopt.h>
#include "common.h"

#define MAX_BUF 4096

struct SquidInfo {
  char protocol[MAX_BUF];
  char domain[MAX_BUF];
  int  dot;  /* true if domain is in dot notation */
  char url[MAX_BUF];
  char orig[MAX_BUF];
  char surl[MAX_BUF];
  char furl[MAX_BUF];
  char *strippedurl;
  int  port;
  char src[MAX_BUF];
  char srcDomain[MAX_BUF];
  char ident[MAX_BUF];
  char method[MAX_BUF];
  char realident[MAX_BUF];
  unsigned int dynamic;
};

/* Converts a hex character to its integer value */
char from_hex(char ch) {
  return isdigit(ch) ? ch - '0' : tolower(ch) - 'a' + 10;
}

/* Converts an integer value to its hex character*/
char to_hex(char code) {
  static char hex[] = "0123456789abcdef";
  return hex[code & 15];
}

/* Returns a url-encoded version of str */
/* IMPORTANT: be sure to free() the returned string after use */
char *url_encode(char *str) {
  char *pstr = str, *buf = malloc(strlen(str) * 3 + 1), *pbuf = buf;
  while (*pstr) {
    if (isalnum(*pstr) || *pstr == '-' || *pstr == '_' || *pstr == '.' || *pstr == '~') 
      *pbuf++ = *pstr;
    else if (*pstr == ' ') 
      *pbuf++ = '+';
    else 
      *pbuf++ = '%', *pbuf++ = to_hex(*pstr >> 4), *pbuf++ = to_hex(*pstr & 15);
    pstr++;
  }
  *pbuf = '\0';
  return buf;
}


int parseLine(char *line, struct SquidInfo *s) {
  char *p, *d = NULL, *a = NULL, *e = NULL, *o, *field;
  int i = 0;
  char c;
  int report_once = 1;
  size_t strsz;
  int ndx = 0;
  
  field = strtok(line,"\t ");
  /*field holds each fetched url*/
  /* Let's first decode the url and then test it. Fixes bug2. */
  HTUnEscape(field);

  if(field == NULL)
    return 0;
  strcpy(s->orig,field);
  /* Now convert url to lowercase chars */ 
  for(p=field; *p != '\0'; p++) {
    *p = tolower(*p);
  }
  s->url[0] = s->protocol[0] = s->domain[0] = s->src[0] = s->ident[0] = 
    s->method[0] = s->srcDomain[0] = s->surl[0] =  '\0';
  s->dot = 0;
  s->port = 0;
  p = strstr(field,"://");
  if(p == NULL) { /* no protocol, defaults to http */
    strcpy(s->protocol,"unknown");
    p = field;
  } else {
    strncpy(s->protocol,field,p - field);
    *(s->protocol + ( p - field)) = '\0';
    p+=3; /* JMC -- 3 == strlen("://") */
    /* Now p only holds the pure URI */
    /* Fix for multiple slash vulnerability (bug1). */
    /* Check if there are still two or more slashes in sequence which must not happen */
    strsz = strlen(p);

    /* loop thru the string 'p' until the char '?' is hit or the "end" is hit */
    while('?' != p[ndx] && '\0' != p[ndx])
    {
        /* in case this is a '://' skip over it, but try to not read past EOS */
        if(3 <= strsz-ndx) {
          if(':' == p[ndx] && '/' == p[ndx+1] && '/' == p[ndx+2]) {
           ndx+=3; /* 3 == strlen("://"); */
          }
        }
        
       /* if this char and the next char are slashes,
 *           then shift the rest of the string left one char */
       if('/' == p[ndx] && '/' == p[ndx+1]) {
         size_t sz = strlen(p+ndx+1);
         strncpy(p+ndx,p+ndx+1, sz);
         p[ndx+sz] = '\0';
          if(1 == report_once) {
            // sgLogError("Warning: Possible bypass attempt. Found multiple slashes where only one is expected: %s", s->orig); 
            report_once--;
          }
      }
      else
      {
        /* increment the string indexer */
	assert(ndx < strlen(p));
        ndx++;
      }
    }
  }

  i=0;
  d = strchr(p,'/'); /* find domain end */
  /* Check for the single URIs (d) */
  e = d;
  a = strchr(p,'@'); /* find auth  */
  if(a != NULL && ( a < d || d == NULL)) 
    p = a + 1;
  a = strchr(p,':'); /* find port */;
  if(a != NULL && (a < d || d == NULL)){
    o = a + strspn(a+1,"0123456789") + 1;
    c = *o;
    *o = '\0';
    s->port = atoi(a+1);
    *o = c;
    e = a;
  }
  o=p;
  strcpy(s->furl,p);
  if (p[0] == 'w' || p[0] == 'f' ) {
    if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||
	(p[0] == 'w' && p[1] == 'e' && p[2] == 'b') ||
	(p[0] == 'f' && p[1] == 't' && p[2] == 'p')) {
      p+=3;
      while (p[0] >= '0' && p[0] <= '9')
	p++;
      if (p[0] != '.')
	p=o; /* not a hostname */
      else
	p++;
    }
  }
  if(e == NULL){
    strcpy(s->domain,o);
    strcpy(s->surl,p);
  }
  else {
    strncpy(s->domain,o,e - o);
    strcpy(s->surl,p);
    *(s->domain + (e - o)) = '\0';
    *(s->surl + (e - p)) = '\0';
  }
  //strcpy(s->surl,s->domain);
  if(strspn(s->domain,".0123456789") == strlen(s->domain))
    s->dot = 1;
  if(d != NULL)
    strcat(s->surl,d);
  s->strippedurl = s->surl;

  while((p = strtok(NULL," \t\n")) != NULL){
    switch(i){
    case 0: /* src */
      o = strchr(p,'/');
      if(o != NULL){
	strncpy(s->src,p,o-p);
	strcpy(s->srcDomain,o+1);
	s->src[o-p]='\0';
	if(*s->srcDomain == '-') {
	  s->srcDomain[0] = '\0';
	}
      } else {
	strcpy(s->src,p);
      }
      break;
    case 1: /* ident */
      if(strcmp(p,"-")){
	strcpy(s->ident,p);
	for(p=s->ident; *p != '\0'; p++) /* convert ident to lowercase chars */
	  *p = tolower(*p);
      } else
	s->ident[0] = '\0';
      break;
    case 2: /* method */
      strcpy(s->method,p);
      break;
    }
    i++;
  }
  if(s->domain[0] == '\0') {
    LOG_Write(DEBUG,"Domain is NULL: %s", s->orig);
    return 0;
  }
  if(s->method[0] == '\0') {
    LOG_Write(DEBUG,"Method is NULL: %s", s->orig);
    return 0;
  }
  return 1;
}

unsigned int onShutdown = 0;

void CB_Shutdown(int mySignal) {
    if(!onShutdown) {
	onShutdown = 1;
	LOG_Write(NOTICE,"Shutdown on signal %d", mySignal);
	LOG_Close();
	exit(0);
    }
}

int doRegExp(char *pattern, char *data) {
    regex_t *regex;
    int err_no=0,match=0; /* For regerror() */
    regmatch_t *result;
    
    regex = (regex_t *) malloc(sizeof(regex_t));
    memset(regex, 0, sizeof(regex_t));
    if((err_no=regcomp(regex, pattern, 0))!=0) /* Compile the regex */ {
	size_t length; 
	char *buffer;
	length = regerror (err_no, regex, NULL, 0);
	buffer = malloc(length);
	regerror(err_no, regex, buffer, length);
	LOG_Write(ERROR, "RegExp error: %s\n", buffer); /* Print the error */
	free(buffer);
	regfree(regex);
	return -1;
    }
    size_t no_sub = regex->re_nsub+1; /* How many matches are there in a line? */

    if((result = (regmatch_t *) malloc(sizeof(regmatch_t) * no_sub))==0) {
	LOG_Write(ERROR, "Cannot allocate required memory !"); /* Print the error */
	return -1;
    }
    
    if(regexec(regex, data, no_sub, result, 0) == 0) { /* Ok, match ! */
	match++;
    }

    regfree(regex); /* Free the regular expression data structure */
    free(regex);

    LOG_Write(DEBUG,"doRegExp: %s %s = %d\n",pattern,data,match);

    return match;
}

MYSQL *conn,*connb,*connc;
MYSQL_RES *res,*resb,*resc;
MYSQL_ROW row,rowb,rowc;

char buf[MAX_BUF],tmp[MAX_BUF];
struct _Config *Config;

void CB_Alarm(int sig) {
    LOG_Write(NOTICE,"CB_Alarm - Purge old entries (Timeout: %d minutes)\n",Config->purgeTimeout);
    sprintf(tmp,"DELETE FROM DB_Access WHERE TIMESTAMPDIFF(MINUTE,NOW(),AddDate) > %d;",Config->purgeTimeout);
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
    }
}

unsigned int checkTimeRule(int ruleId) {
    time_t t;
    struct tm * timeinfo;
    unsigned int retval=0;
    /* Return 1 if VALID, 0 if NOT */
    
    time(&t);
    timeinfo = localtime(&t);
    
    /*
    tm_sec	seconds after the minute	0-61*
    tm_min	minutes after the hour	0-59
    tm_hour	hours since midnight	0-23
    tm_mday	day of the month	1-31
    tm_mon	months since January	0-11
    tm_year	years since 1900	
    tm_wday	days since Sunday	0-6
    tm_yday	days since January 1	0-365
    tm_isdst	Daylight Saving Time flag
    */
    
    sprintf(tmp,"SELECT t2.ID,t2.Name,t2.fromHour,t2.toHour,t2.isNegate,t2.daysOfWeek FROM RulesTime AS t1 INNER JOIN DB_Time AS t2 ON t1.TimeID=t2.ID WHERE t1.RuleID=%d AND (t2.fromHour < CURTIME() < t2.toHour);",ruleId);
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(connc, tmp)) {
	LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connc));
    } else {
	resc = mysql_use_result(connc);
	while((rowc = mysql_fetch_row(resc)) != NULL) {
	 /* Now check the DAYs... */
	    if(rowc[5][timeinfo->tm_wday] == '1') {
		retval = 1;
	    }
	}
	mysql_free_result(resc);
    }
    LOG_Write(DEBUG,"checkTimeRule(%d) = %d\n",ruleId,retval);
    return retval;
}

int getRealIdent(struct SquidInfo *s) {
    int level=-1;
    /* Return identity of source based on DB_Users */
    sprintf(tmp,"SELECT IdentifiedBy,Value,Username,Type FROM DB_Users;");
    /* 
    Username = Identity of source
    IdentifiedBy = How i have to identify this source (IDENT|IP or range IP)
    Value = A value for this rule
    */
    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
    if(mysql_query(conn, tmp)) {
        LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
        return -1;
    }
    res = mysql_use_result(conn);
    while((row = mysql_fetch_row(res)) != NULL) {
        if((strcasecmp(row[0],"IP") == 0) && (doRegExp(row[1],s->src) > 0)) {
	    /* Ok, matched ! Because this is an IP matching, create unique user for each IP... */
	    sprintf(tmp,"%s_%s",row[2],s->src);
	    strncpy(s->realident,row[2],MAX_BUF); /* Realident riferisce al nome utente reale sul DB */
	    strncpy(s->ident,tmp,MAX_BUF); /* ident al nome univoco */
	    level = atoi(row[3]);
	    s->dynamic = 1;
	    break;
	}
        if((strcasecmp(row[0],"IDENT") == 0) && (doRegExp(row[1],s->ident) > 0)) {
	    /* Ok, matched ! */
	    strncpy(s->ident,row[2],MAX_BUF);
	    strncpy(s->realident,row[2],MAX_BUF);
	    level = atoi(row[3]);
	    s->dynamic = 0;
	    break;
	}
    }
    mysql_free_result(res);
    LOG_Write(DEBUG,"getRealIdent(%s) = %s (%d)\n",s->src,s->ident,level);
    return level;
}

int main(int argc, char **argv) {
    struct SquidInfo squidInfo;
    struct sigaction sact;

    /* Default log file */
    char logfile[255];
    strcpy(logfile,"/tmp/squidtl.log\0");

    /* Default config file */
    char configfile[255];
    strcpy(configfile,"/etc/squidtl/config.xml\0");
    
    char tmpEnc[4096],redirectUrl[MAX_BUF];
    
    int rate=0,logVerbose=NOTICE,rulesId;
    
    int c;

    struct option long_options[] = {
               {"verbose",     no_argument,       &logVerbose, DEBUG},
               {"help",        no_argument,       0, 'h'},
               {"config",      required_argument, 0, 'c'},
               {"log",         required_argument, 0, 'l'},
               {0, 0, 0, 0}
    };
     
    while (1) {
	/* getopt_long stores the option index here. */
	int option_index = 0;
	c = getopt_long (argc, argv, "hc:l:", long_options, &option_index);
	/* Detect the end of the options. */
	if(c == -1) break;
	switch (c) {
	    case 'h':
		printf("Usage: %s [options]\n\n--h -help this help\n--verbose\n--c -config config file path\n-l --log log file path");
		exit(0);
	    case 'c':
		strncpy(configfile,optarg,255);
		break;
	    case 'l':
		strncpy(logfile,optarg,255);
		break;
	    default:
		break;
        }
    }
    
    LOG_Init(logfile,logVerbose);

    if((Config = (struct _Config *)CONFIG_Load(configfile)) == NULL) {
	LOG_Write(FATAL,"Config file error: exiting.");
	return(0);
    }

    /* ALARM Signal */
    sigemptyset(&sact.sa_mask);
    sact.sa_flags = 0;
    sact.sa_handler = CB_Alarm;
    sigaction(SIGALRM, &sact, NULL);

    /* HANGUP Signal */    
    sigemptyset(&sact.sa_mask);
    sact.sa_handler = CB_Shutdown;
    sact.sa_flags = SA_NODEFER | SA_RESTART;
    sigaction(SIGHUP, &sact, NULL);

    alarm(360);  /* Trigger ogni 360 secondi = 5 minuti */

    conn = mysql_init(NULL);
    connb = mysql_init(NULL);
    connc = mysql_init(NULL);

    /* Connect to database */
    if (!mysql_real_connect(conn, Config->mysqlHost, Config->mysqlUser, Config->mysqlPassword, Config->mysqlDb, 0, NULL, 0)) {
	LOG_Write(FATAL,"MySQL Error: %s\n", mysql_error(conn));
	return(0);
    }

    if (!mysql_real_connect(connb, Config->mysqlHost, Config->mysqlUser, Config->mysqlPassword, Config->mysqlDb, 0, NULL, 0)) {
	LOG_Write(FATAL,"MySQL Error: %s\n", mysql_error(connb));
	return(0);
    }

    if (!mysql_real_connect(connc, Config->mysqlHost, Config->mysqlUser, Config->mysqlPassword, Config->mysqlDb, 0, NULL, 0)) {
	LOG_Write(FATAL,"MySQL Error: %s\n", mysql_error(connc));
	return(0);
    }

    LOG_Write(NOTICE,"MySQL(%s:%s) connection OK",Config->mysqlHost,Config->mysqlUser);

    while(!onShutdown) {
	while(fgets(buf, MAX_BUF, stdin) != NULL) {
	    LOG_Write(DEBUG,"IN: %s",buf);
	    if(parseLine(buf,&squidInfo) != 1) {
		LOG_Write(ERROR,"Error parsing squid line: %s",buf);
		puts("");
	    } else {
		int userLevel;
		redirectUrl[0] = '\0';
		LOG_Write(DEBUG,"SquidInfo - P:%s D:%s U:%s O:%s S:%s F:%s\n",squidInfo.protocol, squidInfo.domain, squidInfo.url, squidInfo.orig, squidInfo.surl, squidInfo.furl);
		userLevel = getRealIdent(&squidInfo); /* 0=Disabled, -1=Unknown */
		if(userLevel > 0) {
		    /* Add row to DB_Access */
	    	    mysql_real_escape_string(conn, tmpEnc, squidInfo.orig, strlen(squidInfo.orig));
		    sprintf(tmp,"INSERT INTO DB_Access (Src,Domain,FullUrl,Ident,AddDate) VALUES (\"%s\",\"%s\",\"%s\",\"%s\",NOW());",squidInfo.src,squidInfo.domain,tmpEnc,squidInfo.ident);
		    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		    if(mysql_query(conn, tmp)) {
			LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		    }
		    /* Now get groups owned by user (identified by realident) */
	    	    mysql_real_escape_string(conn, tmpEnc, squidInfo.realident, strlen(squidInfo.realident));
		    sprintf(tmp,"SELECT GroupID FROM UserGroup WHERE UserID='%s';",tmpEnc);
		    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
		    if(mysql_query(conn, tmp)) {
			LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
		    }
		    res = mysql_use_result(conn);
		    while((row = mysql_fetch_row(res)) != NULL) {
			/* Ok, now check all rules in those groups */
			sprintf(tmp,"SELECT t1.ID,t1.Type,t1.Value,t1.Action,t1.MaxHits,t1.TimeSlice,t1.redirectUrl FROM DB_Rules AS t1 INNER JOIN RulesGroup AS t2 ON t2.RuleID=t1.ID WHERE GroupID='%s';",row[0]);
			/* 
			Type = Which type of rule is this ? (DOMAIN|IDENT|IP)
			Value = A value for this rule
			Action = What i have to do when triggered ? (BLOCK)
			MaxHits = How many times...
			TimeSlice = ...in how many minutes ?
			redirectUrl = Where i have to redirect the bad guys ?
			*/
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(connb, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connb));
			}
			resb = mysql_use_result(connb);
			while((rowb = mysql_fetch_row(resb)) != NULL) {
			    /* Check rule validity agains Time rules */
			    if(checkTimeRule(atoi(rowb[0])) > 0) { /* <!-- TIME VALIDITY --> */
				/* === DOMAIN LIMITER === */
				if((strcasecmp(rowb[1],"DOMAIN") == 0) && (doRegExp(rowb[2],squidInfo.domain) > 0)) { /* <!-- DOMAIN --> */
				    /* Ok, now check the value */
				    sprintf(tmp,"SELECT Domain FROM DB_Access WHERE Ident=\"%s\" AND TIMESTAMPDIFF(MINUTE,AddDate,NOW()) < %s;",squidInfo.ident,rowb[5]);
				    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
				    if(mysql_query(connc, tmp)) {
					LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(connc));
				    } else {
					resc = mysql_use_result(connc);
					rate = 0;
					while((rowc = mysql_fetch_row(resc)) != NULL) {
					    if(doRegExp(rowb[2],rowc[0]) > 0) {
						rate++;
					    }
					}
					LOG_Write(DEBUG,"LIMIT for domain %s: %s hits in %s minutes. Ident %s have %d\n",rowb[2],rowb[4],rowb[5],squidInfo.ident,rate);
					if(rate > atoi(rowb[4])) {
					    char *redirUrl, *urlencode;
					    if(strncasecmp(rowb[3],"ALLOW") == 0) {
						/* Allow */
						
					    } else {
						/* Default: BLOCK */
						if(strlen(rowb[6]) < 1) {
						    redirUrl = Config->defaultRedirectURL;
						} else {
						    redirUrl = rowb[6];
						}
						urlencode = url_encode(rowb[2]);
						sprintf(redirectUrl,"%s?rule=domain&action=%s&domain=%s&ident=%s&hits=%d&hitlimit=%s&timelimit=%s",redirUrl,rowb[3],urlencode,squidInfo.ident,rate,rowb[4],rowb[5]);
						free(urlencode);
					    }
					    rulesId = atoi(row[0]);
					}
					mysql_free_result(resc);
				    }
				} /* <!-- /DOMAIN --> */
			    } else { /* <!-- /TIME VALIDITY --> */
				LOG_Write(DEBUG,"Rule %s (%s=%s) ignored due to time rule\n",rowb[0],rowb[1],rowb[2]);
			    }
			}
			mysql_free_result(resb);
		    }
		    mysql_free_result(res);
		} else if(userLevel == 0) { /* Disabled user: always block */
		    LOG_Write(DEBUG,"DISABLED USER: BLOCK\n");
		    sprintf(redirectUrl,"%s?rule=disableduser",Config->defaultRedirectURL);
		} else {
		    /* Unknow user action */
		    LOG_Write(DEBUG,"UNKNOWN USER. Default action: %s\n",Config->defaultUserAction);
		    if(strncasecmp(Config->defaultUserAction,"BLOCK",32) == 0) {
			/* Block ALL unknown users */
			sprintf(redirectUrl,"%s?rule=unknownuser",Config->defaultRedirectURL);
		    }
		    if(strncasecmp(Config->defaultUserAction,"PASS",32) == 0) {
			redirectUrl[0] = '\0';
		    }
		    if(strncasecmp(Config->defaultUserAction,"ADD",32) == 0) {
			if(strlen(squidInfo.ident) > 0) {
			    long userId;
			    // Add user to DB_Users
			    sprintf(tmp,"INSERT IGNORE INTO DB_Users (Username,Type,IdentifiedBy,Value,Note,AddDate) VALUES (\"%s\",\"%s\",\"%s\",\"%s\",\"%s\",NOW());",squidInfo.ident,Config->defaultUserLevel,"IDENT",squidInfo.ident,"Added automatically by system based on unknown user action");
			    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			    if(mysql_query(conn, tmp)) {
				LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
			    }
			    // Bind the new user to DB_Groups
			    sprintf(tmp,"INSERT IGNORE INTO UserGroup (UserID,GroupID) VALUES (\"%s\",\"%d\");",squidInfo.ident,Config->defaultUserGroup);
			    LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			    if(mysql_query(conn, tmp)) {
				LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
			    }
			    LOG_Write(NOTICE,"USER %s (%d) ADDED TO DB",squidInfo.ident,Config->defaultUserGroup);
			} else {
			    LOG_Write(ERROR,"Cannot identify unknown user. Automatic ADD is impossibile !");
			}
		    }
		}
		/* Now check if needs to be redirected */
		if(strlen(redirectUrl) > 0) {
		    if(strncasecmp(squidInfo.orig,redirectUrl,MAX_BUF) == 0) {
			/* Prevent infinite loops */
			puts("");
		    } else {
			/* Oh oh, bad guy...needs to be redirected ! */
			fprintf(stdout,"302:%s\n",redirectUrl);
			LOG_Write(NOTICE,"REDIRECT: 302:%s\n",redirectUrl);
			/* Save event to DB for statistic purposes */
			sprintf(tmp,"INSERT INTO DB_Events (RulesID,Src,Ident,AddDate) VALUES (%d,\"%s\",\"%s\",NOW());",rulesId,squidInfo.src,squidInfo.ident);
			LOG_Write(DEBUG,"MySQL: %s\n",tmp);
			if(mysql_query(conn, tmp)) {
			    LOG_Write(ERROR,"MySQL Error: %s\n",mysql_error(conn));
			}
		    }
		} else {
		    LOG_Write(DEBUG,"ALLOWED: %s\n",squidInfo.domain);
		    puts("");
		}
		break;
	    }
	}
	fflush(stdout);
    }
    mysql_close(connb);
    mysql_close(conn);
}
